/*
 * qmail-destroy: Remove messages from a qmail queue.
 * Copyright (C) 2015 LinuxMagic
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): William Storey <william@linuxmagic.com>
 *
 * $Id: queue-remove.c 19983 2015-10-15 22:39:44Z william $
 *
 * Functions related to removal of messages from the queue.
 */

#include <errno.h>
#include <fcntl.h>
#include "queue-remove.h"
#include "queue-remove-util.h"
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static char * __qr_make_qmail_message_path(const char * const,
    const char * const, const int,
    const unsigned long);
static int __qr_flag_recipients_done(const char * const,
    const size_t);
static bool __qr_flag_recipient_done(const int, const int);

//! remove a message from the queue.
/*!
 * @params char* Path to the queue directory. Typically this will be
 *   /var/qmail/queue
 * @params int Directory split count. This tells us how many directories
 *   are present in queue directories (0, 1, etc). Typically this is
 *   23 and is compiled into qmail (conf-split).
 * @params unsigned long Message id
 *   This is the inode number / filename associated with the message.
 *   Per qmail design it is unique to a message in the queue.
 *   Note this is unsigned long as that is what qmail uses internally.
 *   See fmtqfn() in qmail.
 *
 * @returns int Count of how many recipients flagged. -1 if an error
 *   occurred.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * This function will remove a message from the queue.
 *
 * The method involves flagging each recipient as done in the qmail
 * raw files.
 *
 * The files themselves are still present on disk after a return
 * from this function but will be cleaned up by qmail the next time it
 * runs a cleanup (TODO confirm where/when exactly).
 *
 * This means that the message will still show in the queue but it is
 * removed from the queue in that no further deliveries will
 * be attempted.
 *
 * Specific information about removal:
 * - For each message file containing recipients for the message (one
 *   may be in the local directory/channel and one in remote), open it.
 * - For each recipient in the file (format: T<email>\0T<email>\0...)
 *   replace each T with D to flag the recipient as done.
 * - Close the file!
 *
 * TODO:
 * - Possible race condition: Told to remove message id from the queue.
 *   Between then and now it is delivered and cleaned up. Between clean
 *   up and now a new message injected and set up for delivery. We then
 *   flag another message as done than the one intended.
 * - Separate function to trigger immediate clean up of the files? This
 *   may be done through the pulling of the "trigger" (see qmail/INTERNALS).
 * - Should we have this function check the message is in the queue?
 *   Right now I have that in a separate function below.
 */
int qmail_queue_remove(const char * const queue_path, const int dir_split,
    const unsigned long message_id)
{
    // path to the local recipient addresses file.
    char * local_path = NULL;

    // path to the remote recipient addresses file.
    char * remote_path = NULL;

    // TODO: as a define?
    unsigned int bufsize = 1024;

    int count_flagged_local = 0;
    int count_flagged_remote = 0;

    // parameter checks.
    if (!queue_path || strlen(queue_path) == 0 || dir_split <= 0) {
        QRUERROR(strerror(EINVAL));
        return -1;
    }

    // remove it from local queue if it is there.
    local_path = __qr_make_qmail_message_path(queue_path, "local", dir_split,
        message_id);
    if (!local_path) {
        QRUERROR("Unable to make local message path");
        return -1;
    }
    count_flagged_local = __qr_flag_recipients_done(local_path, bufsize);
    if (count_flagged_local < 0) {
        QRUERROR("Unable to flag local recipients done");
        free(local_path);
        return -1;
    }
    free(local_path);
    local_path = NULL;

    // remove it from remote queue if it is there.
    remote_path = __qr_make_qmail_message_path(queue_path, "remote",
        dir_split, message_id);
    if (!remote_path) {
        QRUERROR("Unable to make remote message path");
        return -1;
    }
    count_flagged_remote = __qr_flag_recipients_done(remote_path, bufsize);
    if (count_flagged_remote < 0) {
        QRUERROR("Unable to flag remote recipients done");
        free(remote_path);
        return -1;
    }
    free(remote_path);
    remote_path = NULL;

    // TODO: overflow?
    return count_flagged_local+count_flagged_remote;
}

//! check if this message is in the queue.
/*!
 * @params char* Path to the queue directory. Typically this will be
 *   /var/qmail/queue
 * @params int Directory split count. This tells us how many directories
 *   are present in queue directories (0, 1, etc). Typically this is
 *   23 and is compiled into qmail (conf-split).
 * @params unsigned long Message id
 *   This is the inode number / filename associated with the message.
 *   Per qmail design it is unique to a message in the queue.
 *   Note this is unsigned long as that is what qmail uses internally.
 *   See fmtqfn() in qmail.
 *
 * @returns int 1 if it is, 0 if it is not, -1 if there is an error
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * this function checks if the mess file exists.
 *
 * it can be useful to know if there is a message with this id in the
 * queue in the first place - to be able to tell the difference between
 * no recipients found and message not in the queue.
 */
int qmail_is_message_in_queue(const char * const queue_path,
    const int dir_split, const unsigned long message_id)
{
    char * mess_path = NULL;

    // parameter checks.
    if (!queue_path || strlen(queue_path) == 0 || dir_split <= 0) {
        QRUERROR(strerror(EINVAL));
        return -1;
    }

    // path to the mess file that should always be there if it's queued.
    mess_path = __qr_make_qmail_message_path(queue_path, "mess", dir_split,
        message_id);
    if (!mess_path) {
        QRUERROR("Unable to make message directory");
        return -1;
    }

    if (qru_file_exists(mess_path)) {
        free(mess_path);
        return 1;
    }

    // may exist but we just don't have access...
    // TODO: qru_file_exists() could tell us this more reliably.
    if (errno == EACCES) {
        QRUERROR("No permission to check if the message exists in the queue! (%s)",
            mess_path);
        free(mess_path);
        return -1;
    }

    free(mess_path);
    return 0;
}

//! build path to a split qmail message related file.
/*!
 * @params char* Path to the queue directory. Typically this will be
 *   /var/qmail/queue
 * @params char* Sub directory the message is in. This may be "local"
 *   or "remote" for example.
 * @params int Directory split count. This tells us how many directories
 *   are present in queue directories (0, 1, etc). Typically this is
 *   23 and is compiled into qmail (conf-split).
 * @params unsigned long Message id
 *   This is the inode number / filename associated with the message.
 *   Per qmail design it is unique to a message in the queue.
 *   Note this is unsigned long as that is what qmail uses internally.
 *   See fmtqfn() in qmail.
 *
 * @returns char* Path to the file or NULL if failure. Heap memory.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * we figure out what directory the message is in.
 *
 * we then build the string.
 */
static char * __qr_make_qmail_message_path(const char * const queue_path,
    const char * const sub_dir, const int dir_split,
    const unsigned long message_id)
{
    char * path = NULL;

    // the directory number.
    unsigned long dir_number = 0;

    // parameter checks.
    if (!queue_path || strlen(queue_path) == 0 || !sub_dir ||
        strlen(sub_dir) == 0 || dir_split <= 0) {
        QRUERROR(strerror(EINVAL));
        return false;
    }

    // figure out what directory it is in.
    dir_number = message_id % (unsigned long) dir_split;

    if (!qru_strcat(&path, queue_path) ||
        !qru_strcat(&path, "/") ||
        !qru_strcat(&path, sub_dir) ||
        !qru_strcat(&path, "/") ||
        !qru_icat(&path, (long int) dir_number) ||
        !qru_strcat(&path, "/") ||
        !qru_icat(&path, (long int) message_id)) {
        QRUERROR(strerror(ENOMEM));
        if (path) {
            free(path);
        }
        return NULL;
    }

    QRUDEBUG("made message path [%s] (%d %lu)", path, dir_split, message_id);
    return path;
}

//! flag recipients in a message file as done.
/*!
 * @params char* Path to the message file. This should be a path to a file
 *   under the local or remote directories.
 *   It may or may not exist.
 * @params ssize_t Size of buffer to use in our reads
 *
 * @returns Int a count of how many flagged. -1 if failure.
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * Check if the message exists. If it doesn't there is nothing to do. This is
 * because there are already no recipients.
 *
 * Open it for read/write. For each recipient in the file (\0 separated)
 * write them as DONE by prefixing them with D.
 *
 * Close the file.
 */
static int __qr_flag_recipients_done(const char * const recipients_path,
    const size_t bufsize)
{
    int fd = -1;
    char * buf = NULL;

    // position in the file.
    int fd_pos = 0;

    ssize_t read_size = 0;
    int buf_index = 0;

    // whether we're in the middle of a recipient address (parsing).
    bool in_addr = false;

    int count_flagged = 0;

    // parameter checks.
    if (!recipients_path || strlen(recipients_path) == 0) {
        QRUERROR(strerror(EINVAL));
        return -1;
    }

    // if it does not exist then there are no recipients to flag.
    if (!qru_file_exists(recipients_path)) {
        // may exist but we just don't have access...
        // TODO: qru_file_exists() could tell us this more reliably.
        if (errno == EACCES) {
            QRUERROR("Permission denied checking for recipients file: %s",
                recipients_path);
            return -1;
        }
        QRUDEBUG("file does not exist, no recipients to update %s",
            recipients_path);
        return 0;
    }

    fd = qru_open(recipients_path, true, O_RDWR);
    if (fd == -1) {
        // TODO: race condition could trigger this... it was
        //   there, then cleaned up
        QRUERROR("Unable to open recipients file %s", recipients_path);
        return -1;
    }

    // buffer to read into.
    buf = calloc(bufsize, sizeof(char));
    if (!buf) {
        QRUERROR(strerror(ENOMEM));
        qru_close(fd);
        return -1;
    }

    // read recipients until we hit EOF. they are \0 separated e.g.,
    // Dwilliam@wizard.ca\0Twilliam@linuxmagic.com\0
    // NOTE: we only want to write the single delivery status byte if
    //   anything (T/D). no need to write anything else.
    // TODO: I don't think it is possible using qru_read() to know if we're
    //   at EOF or an error in this scenario... because the data read in is
    //   binary and contains \0... using read() then...
    //   also we need to know exactly how many bytes we read - we need
    //   to know about \0
    while (1) {
        read_size = read(fd, buf, bufsize);
        if (read_size < 0) {
            QRUERROR("read failure: %s", strerror(errno));
            qru_close(fd);
            free(buf);
            return -1;
        }

        // pull out each recipient address.
        // as we reach the one, flag it done if needed.
        for (buf_index = 0; buf_index < read_size; buf_index++) {
            // if we're not in the middle of an address then we must be at the
            // start. we start with a status flag - T or D.
            if (!in_addr) {
                if (buf[buf_index] != 'T' && buf[buf_index] != 'D') {
                    QRUERROR("Invalid recipient format: recipient is not prefixed with status: %c",
                        buf[buf_index]);
                    qru_close(fd);
                    free(buf);
                    return -1;
                }

                // if it's done then we don't need to do anything.
                if (buf[buf_index] == 'D') {
                    // but we are in an address now of course having seen the
                    // flag at the beginning.
                    in_addr = true;
                    QRUDEBUG("recipient already done at %d+%d", fd_pos, buf_index);
                    continue;
                }

                // flag this address done.
                if (!__qr_flag_recipient_done(fd, fd_pos+buf_index)) {
                    QRUERROR("Unable to flag recipient done at %d+%d",
                        fd_pos, buf_index);
                    qru_close(fd);
                    free(buf);
                    return -1;
                }
                QRUDEBUG("flagged recipient done at %d+%d", fd_pos, buf_index);
                count_flagged++;

                // we're now in the address.
                in_addr = true;
                continue;
            }

            // otherwise we're in an address.
            // addresses stop at NULL byte.
            if (buf[buf_index] == '\0') {
                in_addr = false;
            }
        }
        memset(buf, 0, bufsize);

        fd_pos += read_size;

        // stop when we hit EOF.
        if (read_size < (ssize_t) bufsize) {
            QRUDEBUG("hit eof");
            break;
        }
    }

    free(buf);
    buf = NULL;

    if (!qru_close(fd)) {
        QRUERROR("Unable to close recipients file %s", recipients_path);
        return -1;
    }

    return count_flagged;
}

//! flag a recipient done at the given file position
/*!
 * @params int File to write to.
 * @params int File position. We'll write D at this index.
 *
 * @returns bool Whether successful
 *
 * @author William Storey <william@linuxmagic.com>
 *
 * Find out where we are positioned in the file.
 * Seek to the given offset.
 * Write D to indicate done.
 * Seek back to the original offset.
 */
static bool __qr_flag_recipient_done(const int fd, const int pos_to_write)
{
    int orig_pos = 0;
    char * buf = "D";

    // parameter checks.
    if (fd == -1 || pos_to_write < 0) {
        QRUERROR(strerror(EINVAL));
        return false;
    }

    // find out where we are right now in the file. we'll be moving
    // around so we want to know where to go back to.
    orig_pos = qru_tell(fd);
    if (orig_pos < 0) {
        QRUERROR("Unable to determine file position");
        return false;
    }

    // go to where we need to make an update.
    if (!qru_seek(fd, pos_to_write, SEEK_SET)) {
        QRUERROR("Unable to position for write");
        return false;
    }

    // write what we need to.
    if (!qru_write(fd, buf, 1)) {
        QRUERROR("Unable to write done status");
        return false;
    }

    // reposition to where we were.
    if (!qru_seek(fd, orig_pos, SEEK_SET)) {
        QRUERROR("Unable to position back to original location");
        return false;
    }

    return true;
}

#ifdef _TEST_QR
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>

//! program entry
/*!
 * @returns int 0 success, nonzero failure
 *
 * @author William Storey <william@linuxmagic.com>
 */
int main(const int argc, const char * const * const argv)
{
    (void) argc;
    (void) argv;

    // TODO: make these dynamic... but trying to reduce time spent on
    //   unit tests!
    const char * const queue_path = "/tmp/queuetest";
    const char * const test_file = "/tmp/queuetest-file";
    const int dir_split = 23;
    const char * contents = NULL;
    size_t contents_size = 0;
    char buf[1024];
    memset(buf, 0, 1024);

    char * path = NULL;
    char * path2 = NULL;
    char * local_path = NULL;
    char * remote_path = NULL;
    unsigned long message_id = 0;
    char * expected = NULL;
    int fd = -1;
    int i = 0;

    // test building message path.
    // #1: message id under 23
    message_id = 1;
    path = __qr_make_qmail_message_path(queue_path, "local", dir_split,
        message_id);
    assert(qru_strcat(&expected, queue_path));
    assert(qru_strcat(&expected, "/local/1/1"));
    QRUDEBUG("made path [%s] expect [%s]", path, expected);
    assert(strcmp(path, expected) == 0);
    free(path);
    path = NULL;
    free(expected);
    expected = NULL;

    // #2: message id over 23
    message_id = 1234;
    path = __qr_make_qmail_message_path(queue_path, "local", dir_split,
        message_id);
    assert(qru_strcat(&expected, queue_path));
    assert(qru_strcat(&expected, "/local/15/1234"));
    QRUDEBUG("made path [%s] expect [%s]", path, expected);
    assert(strcmp(path, expected) == 0);
    free(path);
    path = NULL;
    free(expected);
    expected = NULL;

    // test flagging status in a file.
    // make a test file.
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));

    // #1: update first one.
    assert(__qr_flag_recipient_done(fd, 0));

    // go read and make sure it's now D.
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);
    assert(qru_close(fd));

    // #2: try updating the other
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));

    // update first one.
    assert(__qr_flag_recipient_done(fd, strlen("Twilliam@wizard.ca")+1));

    // go read and make sure it's now D.
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Twilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    assert(qru_close(fd));

    // test updating a whole file.
    // #1: large enough bufsize to read all at once

    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 1024) == 2);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // #2: small bufsize...

    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 1) == 2);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // #3: small bufsize...
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 2) == 2);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // #4: invalid format - bad status
    contents = "Swilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 2) == -1);

    // #5: some already D
    contents = "Dwilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 2) == 1);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // #6: single recipient
    contents = "Twilliam@wizard.ca\0";
    contents_size = strlen("Twilliam@wizard.ca")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 2) == 1);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // #7: all recipients done already
    contents = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(test_file, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(__qr_flag_recipients_done(test_file, 2) == 0);

    // go check - all should be D
    fd = qru_open(test_file, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);

    // test top level function where we flag local & remote messages
    // set up our directories.

    // queue directory
    if (!qru_directory_exists(queue_path)) {
        assert(mkdir(queue_path, 0700) == 0);
    }

    // local directory
    assert(qru_strcat(&path, queue_path));
    assert(qru_strcat(&path, "/local"));
    if (!qru_directory_exists(path)) {
        assert(mkdir(path, 0700) == 0);
    }

    // local directories.
    for (i = 0; i < dir_split; i++) {
        assert(qru_strcat(&path2, path));
        assert(qru_strcat(&path2, "/"));
        assert(qru_icat(&path2, i));
        if (qru_directory_exists(path2)) {
            free(path2);
            path2 = NULL;
            continue;
        }
        QRUDEBUG("making path [%s]", path2);
        assert(mkdir(path2, 0700) == 0);
        free(path2);
        path2 = NULL;
    }

    free(path);
    path = NULL;

    // remote directory
    assert(qru_strcat(&path, queue_path));
    assert(qru_strcat(&path, "/remote"));
    if (!qru_directory_exists(path)) {
        assert(mkdir(path, 0700) == 0);
    }

    // remote directories
    for (i = 0; i < dir_split; i++) {
        assert(qru_strcat(&path2, path));
        assert(qru_strcat(&path2, "/"));
        assert(qru_icat(&path2, i));
        if (qru_directory_exists(path2)) {
            free(path2);
            path2 = NULL;
            continue;
        }
        assert(mkdir(path2, 0700) == 0);
        free(path2);
        path2 = NULL;
    }

    free(path);
    path = NULL;

    // #1: sample message with local & remote parts
    // local part
    message_id = 123;
    local_path = __qr_make_qmail_message_path(queue_path, "local", dir_split,
        message_id);
    assert(local_path != NULL);
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(local_path, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    // remote part
    remote_path = __qr_make_qmail_message_path(queue_path, "remote",
        dir_split, message_id);
    assert(remote_path != NULL);
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(remote_path, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    // remove them
    assert(qmail_queue_remove(queue_path, dir_split, message_id) == 4);

    // check them.
    fd = qru_open(local_path, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);
    assert(qru_close(fd));

    fd = qru_open(remote_path, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);
    assert(qru_close(fd));

    // #2: sample message with neither local nor remote parts
    assert(unlink(local_path) == 0);
    assert(unlink(remote_path) == 0);
    assert(qmail_queue_remove(queue_path, dir_split, message_id) == 0);
    assert(!qru_file_exists(local_path));
    assert(!qru_file_exists(remote_path));

    // #3: sample message with only local part
    contents = "Twilliam@wizard.ca\0Twilliam@linuxmagic.com\0";
    contents_size = strlen("Twilliam@wizard.ca")+1
        +strlen("Twilliam@linuxmagic.com")+1;
    fd = qru_open(local_path, true, O_CREAT | O_RDWR | O_TRUNC);
    assert(fd != -1);
    assert(qru_write(fd, (char*) contents, contents_size));
    assert(qru_close(fd));

    assert(qmail_queue_remove(queue_path, dir_split, message_id) == 2);

    // check
    fd = qru_open(local_path, false, O_RDONLY);
    assert(fd != -1);
    assert(qru_seek(fd, 0, SEEK_SET));
    assert(qru_read(fd, buf, 1024));
    expected = "Dwilliam@wizard.ca\0Dwilliam@linuxmagic.com\0";
    assert(memcmp(buf, expected, contents_size) == 0);
    memset(buf, 0, 1024);
    assert(qru_close(fd));

    assert(!qru_file_exists(remote_path));

    free(local_path);
    local_path = NULL;
    free(remote_path);
    remote_path = NULL;

    QRUINFO("Tests succeeded");
    return 0;
}

#endif
